/*
 * main.c
 * ======
 *
 * This is the interface to the sff extraction routines (just a basic
 * command line tool).
 *
 * Copyright (C) 2003, 2004, 2005 Osuna Richert Christophe
 *
 *  This file is part of Sffextract.
 *
 *  Sffextract is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * Contributors:
 * _ Winane (important bugfixes and enhancements)
 * _ [E] (hexadecimal output option)
 * _ Dark Saviour (Borland compatibility)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#if defined(__MWERKS__)
/* Note that all references to __MWERKS__ (Metrowerks C) throughout
   Sffextract's source files are based on the documentation provided
   on Metrowerk's website, and are completely untested.
 */
#include <extras.h>
#endif
#include "sffextract.h"


/* ------------------------------------------------------------------------- */
/* Command line help */
char * about = "sffextract version 0.93, Copyright (C)2003, 2004, 2005 Osuna Richert \
Christophe\n\
Contributors: Winane, [E], Dark Saviour\n\
Sffextract comes with ABSOLUTELY NO WARRANTY.\n\
This is free software, and you are welcome to redistribute it under certain\n\
conditions. See GPL (in the accompanying COPYING.txt file) for details.\n\n";

char * usage = "Usage:  sffextract [options] sff_file\n\
where options can be :\
\n\t-d: create subdirectory for PCX files\
\n\t-p: use absolute paths\
\n\t-8: use 8.3 (instead of 10.3) decimal PCX file names (use with caution)\
\n\t-x: use hexadecimal PCX file names\
\n\t-1: override default shared palette (1st vs previous) selection method\
\n\t-n: interpret SFF as post-2001.04.14 character\
\n\t-#: disable palette type specification in text output file\
\n\t-f: do not stop on errors\
\n\t-i: create PCX files\
\n\t-o file: save text output into file\
\n\t-w file: save diagnostic messages into file\
\n\t-a file: apply palette from ACT or PCX file to output PCX files\
\n\t-b: display debug information\
\n\t-v: display program version\
\n\t-h: display this help screen\
\n";
/*  Not currently implemented:
\n\t-r: create PCX files, but do not overwrite existing PCX files\
*/

/* SFF filename */
char sff_filename[FILENAME_MAX] = "\0";

/* output file */
char outputfile[FILENAME_MAX] = "\0";
char use_outputfile = 0;

/* Errors, warnings, notices, and debug info file */
char errorfile[FILENAME_MAX] = "\0";
char use_errorfile = 0;


/* Here we'll decide which string comparison function to use.
   We'll need a case-insensitive function on Microsoft OSes
   (and on pre-OS X Macs, not that Sffextract is currently
   designed to support those, though help there would be welcome),
   but good old ANSI-compliant strcmp will do nicely when
   dealing with case-sensitive file systems such as on Linux.
   Please let the authors know if this part needs revision
   in order to work with other compilers.
   Edit 11/27/05:
   Actually, even in Linux, we might be writing to a case-
   insensitive file system (especially if we're on Knoppix),
   so let's stick with a case-insensitive string compare
   whenever we have one.  -Winane
 */
#if defined(_MSC_VER) || defined(__BORLANDC__) || defined(__MWERKS__)
  #define strcompare stricmp
#elif defined(__GCC__) || defined(__GNUC__) || defined(GNUWIN32)
  #define strcompare strcasecmp
#else
  #define strcompare strcmp
#endif


/* ------------------------------------------------------------------------- */

/* Parse command line
 * returns 1 on error
 */
int parse_cmd_line( int argc, char * argv[] );

#if defined(__GCC__) || defined(__GNUC__)
  #if !defined(__MINGW32__) && !defined(__DJGPP__)
/* This line is to get rid of compile-time warnings with GCC: */
extern char *canonicalize_file_name (const char *);
  #endif
#endif


/* main */
int main( int argc, char * argv[] ) {
  FILE * f; /* output file */
  FILE * erf; /* error messages file */

  /* Full file path names */
#if defined(__MSDOS) || defined(_WIN32) || defined(_MAC)
  char sff_fullp[FILENAME_MAX] = "\0";
  char output_fullp[FILENAME_MAX] = "\0";
  char error_fullp[FILENAME_MAX] = "\0";
#define free_paths 
#else
  char *sff_fullp = NULL;
  char *output_fullp = NULL;
  char *error_fullp = NULL;
#define free_paths free(sff_fullp);free(output_fullp);free(error_fullp)
#endif


  set_output_file( stdout );
  set_error_file( stderr );
  set_display_header( TRUE );
  set_display_contents( TRUE );
  set_display_outfile( TRUE );

    if(! parse_cmd_line( argc, argv ) ){

        /* There doesn't seem to be an ANSI-standard way of obtaining full paths.
           To do that, Microsoft provides the _fullpath function, DJGPP provides
           the _fixpath function, GNU provides the canonicalize_file_name
           function, and POSIX provides the realpath function.  If none of these
           work with your compiler, or if the pre-processor directives need to
           be modified to select the one that your compiler supports, then
           please inform the authors.  (Confirmation of proper functionality on
           other compilers would also be appreciated.)
           Note that Metrowerks MSL only provides their _fullpath function in
           Windows, and I didn't find any alternative for POSIX, Mac OS, or Mac
           OS X in the documentation they provide online.  -Winane
         */
#if defined(_MSC_VER) || defined(__MINGW32__) || defined(__BORLANDC__) || defined(__MWERKS__)
        if( use_outputfile )
             _fullpath(output_fullp, outputfile, FILENAME_MAX);
        if( use_errorfile )
             _fullpath(error_fullp, errorfile, FILENAME_MAX);
        _fullpath(sff_fullp, sff_filename, FILENAME_MAX);
#elif defined(__DJGPP__)
        if( use_outputfile )
            _fixpath( outputfile, output_fullp );
        if( use_errorfile )
            _fixpath( errorfile, error_fullp );
        _fixpath( sff_filename, sff_fullp );
#elif defined(__GCC__) || defined(__GNUC__)
        /* Annoyingly enough, canonicalize_file_name and realpath both return
           NULL if the file whose name is being checked doesn't exist yet.  When
           this is the case, we'll just stick with the file name as given by the
           user, and hope that anyone using Linux or Unix would be sensible
           enough not to specify two names for the two output files that resolve
           to the same location.  Please let the authors know if you have a
           better solution.  -Winane
        */
        if( use_outputfile ){
            output_fullp = canonicalize_file_name(outputfile);
            if( output_fullp == NULL ) output_fullp = strdup(outputfile);
        }
        if( use_errorfile ){
            error_fullp = canonicalize_file_name(errorfile);
            if( error_fullp == NULL ) error_fullp = strdup(errorfile);
        }
        sff_fullp = canonicalize_file_name(sff_filename);
#else
        if( use_outputfile ){
            output_fullp = realpath(outputfile, NULL);
            if( output_fullp == NULL ) output_fullp = strdup(outputfile);
        }
        if( use_errorfile ){
            error_fullp = realpath(errorfile, NULL);
            if( error_fullp == NULL ) error_fullp = strdup(errorfile);
        }
        sff_fullp = realpath(sff_filename, NULL);
#endif
        if( use_outputfile && (output_fullp == NULL) ){
            fprintf( stderr, "Error determining full path name of output file.\n" );
            free_paths;
            return 1;
        }
        if( use_errorfile && (error_fullp == NULL) ){
            fprintf( stderr, "Error determining full path name of error file.\n" );
            free_paths;
            return 1;
        }
        if( sff_fullp == NULL ){
            fprintf( stderr, "Error determining full path name of SFF file.\n" );
            free_paths;
            return 1;
        }


        if( use_outputfile && use_errorfile ){

            if( !strcompare( output_fullp, error_fullp ) ){
                fprintf( stderr, "Error: Text data filename must be different from diagnostic messages filename.\n" );
                free_paths;
                return 1;
            }
        }


        if( use_outputfile ){

            if( !strcompare( output_fullp, sff_fullp ) ){
                fprintf( stderr, "Error: Text data filename must be different from SFF filename.\n" );
                free_paths;
                return 1;
            }

            if( (f = fopen( outputfile, "w" )) == NULL ){
                fprintf( stderr, "Error: Could not open output file for writing: %s", outputfile );
                free_paths;
                return 1;
            }
            else set_output_file( f );
        }


        if( use_errorfile ){
            if( !strcompare( error_fullp, sff_fullp ) ){
                fprintf( stderr, "Error: Diagnostic messages filename must be different from SFF filename.\n" );
                free_paths;
                return 1;
            }
            free_paths;

            if( (erf = fopen( errorfile, "w" )) == NULL ){
                fprintf( stderr, "Error: Could not open message output file for writing: %s", errorfile );
                return 1;
            }
            else set_error_file( erf );

            if( extract( sff_filename ) ){
                fprintf( stderr, "Aborting due to error encountered while processing SFF file.\n\
                    Check diagnostic message file for details.\n" );
                return 1;
            }
            else{
                return 0;
            }
        }

        else{
            free_paths;
            return extract( sff_filename );
        }
    }
    else return 0;
}


/* ------------------------------------------------------------------------- */
/* Parse command line.
 * returns 1 on error
 */
int parse_cmd_line( int argc, char * argv[] ) {
    int state = 1;
    int parsing, filename_exists = FALSE;
    int currargc = 1, currargv, length;

    if( argc > 1 ) {
        length = strlen(argv[1]);
        parsing = TRUE;
    }
    else parsing = FALSE;

    while( parsing ) {
        switch( state ) {

        case 0 : /* exiting */
            parsing = FALSE;
            break;



        case 1 : /* main */
            if( argv[currargc][0] == '-' ) {
                currargv = 1;
                state = 2;
            }
            /* New change as of version 0.8 */
            else if( argv[currargc][0] == '>' ){
                state = 0;
            }
            else if( length > 1 ){
                if( argv[currargc][1] == '>' ){
                    state = 0;
                }
                else state = 3;
            }
            else state = 3;
            break;



        case 2 : /* reading options */
            if( currargv == length ) { /* finished argument */
              state = 7;
            }
            else { /* keep on reading */
              switch( argv[currargc][currargv] ) {
                case 'f' : set_force( TRUE ); break;
                case 'i' : set_create_pcx( TRUE ); break;
              /*case 'r' : set_create_pcx( TRUE );
                    set_no_overwrite( TRUE ); break;*/
                case 'b' : set_debug( TRUE ); break;
                case 'o' : state = 4; break;
                case 'd' : state = 5; break;
                /* new , by [E] 13/08/2003 */
                case 'x' : set_hex_output( TRUE ); break;
                /* ends, by [E] 13/08/2003 */
                case '8' : set_short_output( TRUE ); break;
                case 'p' : set_absolute_paths( TRUE ); break;
                case 'a' : state = 6; break;
                case '1' : set_invert_shared( TRUE ); break;
                case 'n' : set_linux_kyara( TRUE ); break;
                case '#' : set_display_options( FALSE ); break;
                case 'w' : state = 8; break;
                case 'v' : fprintf( stderr, about ); return 1;
                case 'h' : fprintf( stderr, about ); fprintf( stderr, usage );
                       return 1;
                default : fprintf( stderr, about ); fprintf( stderr, usage );
                      return 1;
              }
              currargv++;
            }
            break;


        case 3 : /* read sff filename */
            if( strlen( argv[currargc] ) >= FILENAME_MAX ){
                fprintf( stderr, "Error: Specified SFF file name is too long.\n" );
                return 1;
            }
            strcpy( sff_filename, argv[currargc] );
            filename_exists = TRUE;
            state = 7;
            break;


        case 4: /* read output filename */
            if( currargc == (argc-1) ) {
                fprintf( stderr, about );
                fprintf( stderr, usage );
                return 1;
            }
            use_outputfile = 1;
            if( strlen( argv[++currargc] ) >= FILENAME_MAX ){
                fprintf( stderr, "Error: Specified output file name is too long.\n" );
                return 1;
            }
            strcpy( outputfile, argv[currargc] );
            state = 7;
            break;


        case 5 : /* read directory name */
            set_subdirectory( TRUE );
            /* the following will be added later */
            /*
            if( currargc == (argc-1) ) {
              fprintf( stderr, about );
              fprintf( stderr, usage );
              return 1;
            }
            set_output_dir( argv[++currargc]);
            */
            state = 7;
            break;


        case 6 : /* read act filename */
            if( currargc == (argc-1) ) {
                fprintf( stderr, about );
                fprintf( stderr, usage );
                return 1;
            }
            if( strlen( argv[++currargc] ) >= FILENAME_MAX ){
                fprintf( stderr, "Error: Specified palette file name is too long.\n" );
                return 1;
            }
            set_act_file(argv[currargc]);
            set_use_act( TRUE );
            set_create_pcx( TRUE ); /* Not sure if this should be done... */
            state = 7;
            break;


        case 7 : /* exit test */
            if( ++currargc == argc )
                state = 0; /* finished */
            else {
                state = 1; /* back to main state */
                length = strlen( argv[currargc] );
            }
            break;


        case 8: /* read warnings file name */
            if( currargc == (argc-1) ) {
                fprintf( stderr, about );
                fprintf( stderr, usage );
                return 1;
            }
            if( strlen( argv[++currargc] ) >= FILENAME_MAX ){
                fprintf( stderr, "Error: Specified error file name is too long.\n" );
                return 1;
            }
            use_errorfile = 1;
            strcpy( errorfile, argv[currargc] );
            state = 7;
            break;

        }
    }

    if( filename_exists ) return 0;
    else {
        fprintf( stderr, about );
        fprintf( stderr, usage );
        return 1;
    }
}
